/********************************************************************************/
/* HALF-LIFE MODEL VALIDATOR v1.0												*/
/*																				*/
/*	Written by David "Nighthawk" Flor ( dflor@mach3.com )						*/
/*	(c) 2000, Mach III Enterprises												*/
/*		http://www.planethalflife.com/mach3/									*/
/*																				*/
/*	Designed for use in "The Opera" 											*/
/*		http://www.halflife.net/opera/											*/
/*																				*/	
/********************************************************************************/
/*	Snippets of code taken from the "mdlviewer" project in the Half-Life SDK	*/
/*																				*/	
/*	Copyright (c) 1999, Valve LLC. All rights reserved.							*/
/*																				*/	
/*	This product contains software technology licensed from Id 					*/
/*	Software, Inc. ("Id Technology").  Id Technology (c) 1996 Id Software, Inc. */
/*	All Rights Reserved.														*/
/*																				*/	
/*   Use, distribution, and modification of this source code and/or resulting	*/
/*   object code is restricted to non-commercial enhancements to products from	*/
/*   Valve LLC.  All other use, distribution, or modification is prohibited		*/
/*   without written permission from Valve LLC.									*/
/********************************************************************************/

/********************************************************************************/
/********************************************************************************/
/*	NOTE: You can pretty much use this anywhere you want; all that I ask is		*/
/*	that credit be given where it's due. - DFlor								*/
/********************************************************************************/
/********************************************************************************/

/********************************************************************************/
/*	KNOWN ISSUES:																*/
/*	- PAK files: for now, this can't open them. In this version, at least until	*/
/*	  I can get a PAK file decompressor or something, simply don't use PAK		*/
/*	  files.																	*/
/*  - Not all models and animation types were checked. It does work on the		*/
/*    Half-Life player model, though!											*/
/********************************************************************************/

/********************************************************************************/
/*	The following code is designed for my current use in "The Opera"; all that	*/
/*	would be needed is changing a few lines (listed below).						*/
/*																				*/
/*	This is the most rudimentary form of this, and has some room for 			*/
/*	improvement.																*/
/********************************************************************************/
/*	NOTE: This can, in theory, also be used on the server, but I can't really	*/
/*	think of a practical reason as to why it would be necessary there.			*/
/********************************************************************************/
	
#include	"hud.h"
#include	"util.h"
#include	"parsemsg.h"
#include	"../engine/studio.h"

// Change to be your game directory
#define				kGameDirectory				"opera"	 /***** CHANGE THIS!!! *****/

// If you have more than 128 models, I'd really be surprised!
#define				MAX_MODEL_ENTRIES			128

/********************************************************************************/
class CModelEntry
{
public:
	CModelEntry()		{	Clear();	}

	const char			*szModelName;
	Vector				vMinBounds;
	Vector				vMaxBounds;
	long				iVertexCount;

	Vector				vMinBone;
	Vector				vMaxBone;
	long				iBoneCount;

	void				ReadModel( const char *szModelName );

	void				AddVertex( const Vector &vPoint );
	void				AddBone( const Vector &vPoint );
	void				AddBonesToVertices( void );

	void				Clear( void );	 // Force of habit...
};

static const char *gModelListing[] =
{
/*****
	This is an array of all the model files that are going to be validated. 
	*** Make sure the last one is NULL!!!
*****/
	"models/temp/player.mdl",		/***** CHANGE THIS!!! *****/

	NULL
};

static CModelEntry gModelEntries[ MAX_MODEL_ENTRIES ];
static long gModelEntryCount = 0;

//-----------------------------------------------------------------------------
void CModelEntry::Clear( void )
{
	szModelName = NULL;
	vMinBounds = vMaxBounds = Vector( 0, 0, 0 );
	iVertexCount = 0;

	vMinBone = vMaxBone = Vector( 0, 0, 0 );
	iBoneCount = 0;
}

void CModelEntry::AddVertex( const Vector &vPoint )
{
	if (iVertexCount == 0)
	{
		vMinBounds = vMaxBounds = vPoint;
	}
	else
	{
		vMinBounds.x = min( vMinBounds.x, vPoint.x );
		vMinBounds.y = min( vMinBounds.y, vPoint.y );
		vMinBounds.z = min( vMinBounds.z, vPoint.z );

		vMaxBounds.x = max( vMaxBounds.x, vPoint.x );
		vMaxBounds.y = max( vMaxBounds.y, vPoint.y );
		vMaxBounds.z = max( vMaxBounds.z, vPoint.z );
	}
	iVertexCount++;
}

void CModelEntry::AddBone( const Vector &vPoint )
{
	if (iBoneCount == 0)
	{
		vMinBone = vMaxBone = vPoint;
	}
	else
	{
		vMinBone.x = min( vMinBone.x, vPoint.x );
		vMinBone.y = min( vMinBone.y, vPoint.y );
		vMinBone.z = min( vMinBone.z, vPoint.z );

		vMaxBone.x = max( vMaxBone.x, vPoint.x );
		vMaxBone.y = max( vMaxBone.y, vPoint.y );
		vMaxBone.z = max( vMaxBone.z, vPoint.z );
	}
	iBoneCount++;
}

/*****
void CModelEntry::AddBonesToVertices( void );

	This normalizes all the bone positions.

	There are some animations (long jump, some death sequences, etc...) that
	have a "lateral movement" value as part of the animation. This is not
	deducted from the animation's bounding area, so the bounding area might
	represent a larger area than the animation really is.

	What the below does is create a bounding area based on where the BONE
	ORIGINS actually are. It doesn't take the position of attached vertices
	in to consideration because, if they were an issue, they would be caught
	by the standard refrence frame vertex check.
*****/
void CModelEntry::AddBonesToVertices( void )
{
	Vector vAdjust;

	if (iBoneCount > 0)
	{
		vAdjust = (vMaxBone - vMinBone);

		AddVertex( vAdjust / 2 );
		AddVertex( -(vAdjust / 2) );

		vMinBone = vMaxBone = Vector( 0, 0, 0 );
		iBoneCount = 0;
	}
}

//-----------------------------------------------------------------------------
/*****	
static void CalcBonePosition( int frame, 
							  mstudiobone_t *pbone, 
							  mstudioanim_t *panim, 
							  float *pos );

	Ripped from Valve's "mdlviewer". Need to determine a bone's postiion
	in space!
*****/
static void CalcBonePosition( int frame, 
							  mstudiobone_t *pbone, 
							  mstudioanim_t *panim, 
							  float *pos )
{
	float				s = 0;	// Don't care about interpolation - DFlor
	int					j, k;
	mstudioanimvalue_t	*panimvalue;

	for (j = 0; j < 3; j++)
	{
		pos[j] = pbone->value[j]; // default;
		if (panim->offset[j] != 0)
		{
			panimvalue = (mstudioanimvalue_t *)((byte *)panim + panim->offset[j]);
			
			k = frame;
			// find span of values that includes the frame we want
			while (panimvalue->num.total <= k)
			{
				k -= panimvalue->num.total;
				panimvalue += panimvalue->num.valid + 1;
			}
			// if we're inside the span
			if (panimvalue->num.valid > k)
			{
				// and there's more data in the span
				if (panimvalue->num.valid > k + 1)
				{
					pos[j] += (panimvalue[k+1].value * (1.0 - s) + s * panimvalue[k+2].value) * pbone->scale[j];
				}
				else
				{
					pos[j] += panimvalue[k+1].value * pbone->scale[j];
				}
			}
			else
			{
				// are we at the end of the repeating values section and there's another section with data?
				if (panimvalue->num.total <= k + 1)
				{
					pos[j] += (panimvalue[panimvalue->num.valid].value * (1.0 - s) + s * panimvalue[panimvalue->num.valid + 2].value) * pbone->scale[j];
				}
				else
				{
					pos[j] += panimvalue[panimvalue->num.valid].value * pbone->scale[j];
				}
			}
		}
	}
}

/*****
void CModelEntry::ReadModel( const char *szModelName );

	This function does all the work: reads the ENTIRE MDL file in to memory,
	then runs through it, vertex by vertex, tabulating the model bounds.
*****/
void CModelEntry::ReadModel( const char *szModelName )
{
	FILE		*fp;
	long		lSize, lRead;
	char		*pBuffer;
	char		*pTempBuffer;
	char		szFullName[ MAX_PATH ];
	float		flBone[3];

	if (strchr( szModelName, ':' ))
		strcpy( szFullName, szModelName );
	else
		sprintf( szFullName, "%s/%s", kGameDirectory, szModelName );
	fp = fopen( szFullName, "rb" );
	if (fp)
	{
		fseek( fp, 0, SEEK_END );
		lSize = ftell( fp );
		fseek( fp, 0, SEEK_SET );

		pBuffer = (char *)malloc( lSize );
		if (pBuffer)
		{
			lRead = fread( pBuffer, 1, lSize, fp );
			if (lRead == lSize)
			{
				studiohdr_t			*pHeader;
				mstudiobodyparts_t	*pBodyParts;
				mstudiomodel_t		*pModel;
				long				iCnt;
				long				iModelCnt, iModels;
				long				iVert, iVertCnt;
				vec3_t				*pVert;
				mstudiobone_t		*pBone;
				long				iBoneCnt, iBones;
				mstudioseqdesc_t	*pSequence;
				long				iSequenceCnt, iSequences;
				mstudioanim_t		*pAnim;
				long				iFrameCnt, iFrames;

				pHeader = (studiohdr_t *)pBuffer;

				pTempBuffer = (pBuffer + pHeader->bodypartindex);
				pBodyParts = (mstudiobodyparts_t *)pTempBuffer;
				iModels = 0;
				for (iCnt = 0; iCnt < pHeader->numbodyparts; iCnt++)
					iModels += pBodyParts[iCnt].nummodels;
				pTempBuffer += (pHeader->numbodyparts * sizeof(mstudiobodyparts_t));

			//***************************************************************
			//*** Pound through the base reference frame vertices
			//***************************************************************
				pModel = (mstudiomodel_t *)pTempBuffer;
				for (iModelCnt = 0; iModelCnt < iModels; iModelCnt++)
				{
					iVert = pModel[iModelCnt].numverts;
					pVert = (vec3_t *)(pBuffer + pModel[iModelCnt].vertindex);
					for (iVertCnt = 0; iVertCnt < iVert; iVertCnt++)
					{
						AddVertex( pVert[iVertCnt] );
					}
				}

			//***************************************************************
			//*** Pound through the animations and determine where the 
			//*** bones are. Necessary because, in theory, someone could
			//*** modify the animations and add extra bones!
			//***************************************************************
				pBone = (mstudiobone_t *)(pBuffer + pHeader->boneindex);
				iBones = pHeader->numbones;

				pSequence = (mstudioseqdesc_t *)(pBuffer + pHeader->seqindex);
				iSequences = pHeader->numseq;
				for (iSequenceCnt = 0; iSequenceCnt < iSequences; iSequenceCnt++)
				{
					iFrames = pSequence[iSequenceCnt].numframes;

					pTempBuffer = (pBuffer + pSequence[iSequenceCnt].animindex);
					pAnim = (mstudioanim_t *)pTempBuffer;
					for (iBoneCnt = 0; iBoneCnt < iBones; iBoneCnt++)
					{
						for (iFrameCnt = 0; iFrameCnt < iFrames; iFrameCnt++)
						{
							CalcBonePosition( iFrameCnt, pBone + iBoneCnt,
								pAnim, flBone );

							AddBone( flBone );
						}
					}

					AddBonesToVertices();
				}

			}
			free( pBuffer );
		}
		fclose( fp );
	}
}

/*****
void ClearModelListing( void );

	Blanks out the model listing, forcing them all to be re-read from file the
	next time "ValidateModels" is called
*****/
void ClearModelListing( void )
{
	long iCnt;

	gModelEntryCount = 0;
	for (iCnt = 0; iCnt < MAX_MODEL_ENTRIES; iCnt++)
		gModelEntries[iCnt].Clear();
}

/*****
void ScanModelListing( void );

	Reads in all the model files based on the above filename array
*****/
void ScanModelListing( void )
{
	long		iCnt;

	ClearModelListing();

	iCnt = 0;
	while ( gModelListing[iCnt] )
	{
		gModelEntries[gModelEntryCount].szModelName = gModelListing[iCnt];
		gModelEntries[gModelEntryCount].ReadModel( gModelListing[iCnt] );
		
		if (gModelEntries[gModelEntryCount].iVertexCount > 0)
			gModelEntryCount++;
		else
		{
			gModelEntryCount++;
		}

		iCnt++;
	}
}

/*****
bool ValidateModels( void );

	This function is where it all takes place!

	Right now, this isn't designed very optimally, because it first reads through
	ALL the models, then scans them all. It might be easier to scan them as they
	are read, but I guess I'm lazy that way.

	Where you call this function from is really up to you:
	- Calling it from "DLLMain", and returning FALSE from "DLLMain" if "ValidateModels"
	  is FALSE will cause the client DLL to NOT LOAD, and Half-Life will stop. The
	  problem with this is that there's no stopping someone from changing models while
	  the game is running; when the next map loads, the models are pre-cached again but
	  the DLL is never unloaded and, thus, this function might not get called.

	- From "VidInit". Probably the best place, but I have not confirmed how many times
	  this might be called. Do remember that this is a somewhat expensive process, and
	  should not be called too repeatedly.

	  Using the second method allows for some creativity when it comes to your mod.
	  For instance, you might not want to simply dump the player out of Half-Life,
	  but you might want to torture him a little. If this returns FALSE, you can
	  conceivably send commands to the server which might make his accuracy worse,
	  might make him not increase his score, etc... Client side you can remove his HUD,
	  make it white bright, pitch black, chromatic, etc... Possibilities are endless!
*****/
bool ValidateModels( void )
{
	long	iCnt;
	Vector	vBounds, vMaxBounds;

	if (gModelEntryCount == 0)
		ScanModelListing();

	if (gModelEntryCount == 0)
		return false;

	for (iCnt = 0; iCnt < gModelEntryCount; iCnt++)
	{
		vBounds = gModelEntries[iCnt].vMaxBounds - gModelEntries[iCnt].vMinBounds;

/*****
	The bounds which you consider to be valid for different model types is
	arguable. In certain circumstances, you might actually want a taller player
	model, or a bigger weapon, or whatever... This is fully customizable to your
	heart's content.

	I ultimately chose a size of 100x100x100 because, well, the numbers sounded
	nice and round. Examples of other possible check values are listed.

	Addendum: Actually, the value of 100 works out rather well: the Half-Life
	"long jump" animation is 92 units long, at least according to the above
	calculations...
*****/

/*****
		if ( !strnicmp( gModelEntries[iCnt].szModelName, "models/player", 13 ) )
			vMaxBounds = Vector( 40, 90, 40 );
		else if ( !strnicmp( gModelEntries[iCnt].szModelName, "models/w_", 9 ) )
			vMaxBounds = Vector( 80, 80, 32 );
		else if ( !strnicmp( gModelEntries[iCnt].szModelName, "models/v_", 9 ) )
			vMaxBounds = Vector( 80, 80, 80 );
		else
			vMaxBounds = Vector( 32, 48, 50 );
*****/
		vMaxBounds = Vector( 100, 100, 100 );

		if (vBounds.x > vMaxBounds.x)
			return false;
		if (vBounds.y > vMaxBounds.y)
			return false;
		if (vBounds.z > vMaxBounds.z)
			return false;
	}

	return true;
}

/********************************************************************************/
/* HALF-LIFE MODEL VALIDATOR													*/
/*																				*/
/*	Written by David "Nighthawk" Flor ( dflor@mach3.com )						*/
/* (c) 2000, Mach III Enterprises												*/
/*		http://www.planethalflife.com/mach3/									*/
/*																				*/
/*	Designed for use in "The Opera" 											*/
/*		http://www.halflife.net/opera/											*/
/*																				*/	
/********************************************************************************/
